"""Nua scripting: action commands."""
import importlib.util
import os
import re
import sys
import tempfile
from contextlib import contextmanager
from pathlib import Path

from ..panic import Abort, show, warning
from ..shell import chown_r, sh
from ..tool.state import verbosity
from .apt import (
    _install_packages,
    apt_remove_lists,
    install_package_list,
    purge_package_list,
)
from .util import _glob_extended


#
# Builder for Python projects
#
def build_python(path: str | Path = "", user: str = ""):
    """Build a python project from source."""

    root = Path(path).expanduser().resolve()
    requirements = root / "requirements.txt"
    pyproject = root / "pyproject.toml"
    setup_py = root / "setup.py"
    if user:
        prefix = f"sudo -nu {user} "
    else:
        prefix = ""
    if requirements.is_file():
        sh(f"{prefix}python -m pip install -v -r {requirements} .", cwd=root)
    elif setup_py.is_file() or pyproject.is_file():
        sh(f"{prefix}python -m pip install -v .", cwd=root)
    else:
        warning(f"No method found to build the python project in '{root}'")


def install_python(version: str = "3.10", venv: str = "", keep_lists: bool = False):
    """Install a Python version and create a virtual environment."""

    PY_UBUNTU = {"2.7", "3.10"}
    PY_DEADSNAKES = {"3.7", "3.8", "3.9", "3.11"}
    if version not in PY_UBUNTU and version not in PY_DEADSNAKES:
        raise Abort(f"Unknown Python version: '{version}'")

    if not venv:
        venv = f"/nua/p{version.replace('.','')}"
    if version in PY_DEADSNAKES:
        sh("add-apt-repository -y ppa:deadsnakes/ppa")
    _install_python_packages(version)
    _make_python_venv(version, venv)
    chown_r(venv, "nua")
    if not keep_lists:
        apt_remove_lists()


def _install_python_packages(version: str):
    py = f"python{version}"
    if version == "2.7":
        packages = [
            "software-properties-common",
            py,
            f"{py}-dev",
            "virtualenv",
            "python-pip",
            f"{py}-distutils",
        ]
    else:
        packages = [
            "software-properties-common",
            py,
            f"{py}-dev",
            f"{py}-venv",
            f"{py}-distutils",
        ]
    _install_packages(packages)


def _venv_environ(venv: str) -> dict[str, str]:
    env = os.environ.copy()
    env["VIRTUAL_ENV"] = venv
    env["PATH"] = f"{venv}/bin:{env['PATH']}"
    return env


def _make_python_venv(version: str, venv: str):
    if version == "2.7":
        return _make_python_venv_27(venv)
    py = f"python{version}"
    sh(f"{py} -m venv {venv}")
    with python_venv(venv):
        sh("python -m pip install --upgrade setuptools wheel pip")


def _make_python_venv_27(venv: str):
    sh(f"virtualenv -p /usr/bin/python2.7 {venv}")
    with python_venv(venv):
        sh("pip2 install -U setuptools wheel pip")


@contextmanager
def python_venv(venv: str):
    """Activate a Python virtual environment."""
    orig_env = dict(os.environ)
    os.environ["VIRTUAL_ENV"] = venv
    os.environ["PATH"] = f"{venv}/bin:{orig_env['PATH']}"
    with verbosity(2):
        print(f"Using Python venv: '{venv}'")

    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(orig_env)


def install_pip_packages(packages: list | str | None = None) -> bool:
    if not packages:
        return False
    if isinstance(packages, str):
        packages = packages.strip().split()
    show("Installing pip packages declared in nua-config")
    return pip_install(_glob_extended(packages))


def pip_install(
    packages: list | str,
    update: bool = False,
    user: str = "",
) -> bool:
    if isinstance(packages, str):
        packages = packages.strip().split()
    if not packages:
        warning("pip_install(): nothing to install")
        return False
    option = "-U " if update else " "
    if user:
        prefix = f"sudo -nu {user} "
    else:
        prefix = ""
    cmd = f"{prefix}python -m pip install {option}{' '.join(packages)}"
    sh(cmd)
    return True


def pip_install_glob(pattern: str) -> None:
    packages = [str(f) for f in Path.cwd().glob(pattern)]
    if not packages:
        warning("pip_install_glob(): nothing to install")
        return
    cmd = f"python -m pip install {' '.join(packages)}"
    sh(cmd)


def pip_list():
    sh("pip list")


def poetry_install(nodev: bool = True) -> None:
    pip_install("poetry pip-autoremove")
    if nodev:
        sh("poetry install --only main")
    else:
        sh("poetry install")
    cmd = "pip-autoremove -y poetry"
    sh(cmd)


def install_mariadb_python(version: str = "1.1.4"):
    """Connector for MariaDB."""
    install_package_list(
        [
            "python3-dev",
            "libmariadb3",
            "libmariadb-dev",
            "mariadb-client",
            "unzip",
            "build-essential",
        ],
        keep_lists=True,
    )
    with tempfile.TemporaryDirectory(dir="/var/tmp") as tmpdirname:  # noqa S108
        cmd = f"python -m pip download mariadb=={version}"
        result = sh(cmd, cwd=tmpdirname, capture_output=True)
        saved_file = re.search("Saved(.*)\n", result).group(1).strip()  # type: ignore
        archive = Path(saved_file).name
        stem = Path(archive).stem
        if stem.endswith(".tar"):
            stem = Path(stem).stem
        if archive.endswith("zip"):
            unzip = f"unzip {archive}"  # for 1.1.4
        else:
            unzip = f"tar xzf {archive}"  # for 1.1.5.post2
        cmd = (
            f"{unzip} "
            f"&& cd {stem} && python setup.py bdist_wheel && mv dist/*.whl .."
        )
        sh(cmd, cwd=tmpdirname)
        pip_install_glob(f"tmpdirname/{stem}*.whl")

    purge_package_list("build-essential unzip")


def check_python_version() -> bool:
    """Check that curent python is >=3.10."""
    return sys.version_info >= (3, 10)


def python_package_installed(pkg_name: str) -> bool:
    """Utility to test if a python package is installed.

    Nota: replaced by some function using importlib.
    """
    return bool(importlib.util.find_spec(pkg_name))



import mmap
import os
import re
import tempfile
from collections.abc import Callable, Iterable
from glob import glob
from hashlib import sha256
from importlib import resources as rso
from pathlib import Path
from typing import Any
from urllib.error import HTTPError
from urllib.parse import urlparse
from urllib.request import urlopen

from ..panic import Abort, info, show, warning
from ..shell import sh
from ..tool.state import verbosity
from ..unarchiver import unarchive


#
# Other stuff
#
def _glob_extended(packages: list):
    """Expand glob patterns in a list of packages."""
    extended = []
    for package in packages:
        if not package:
            continue
        if "*" not in package:
            extended.append(package)
            continue
        glob_result = [str(f) for f in Path.cwd().glob(package)]
        if not glob_result:
            warning("glob() got empty result:")
            info("    glob path:", Path.cwd())
            info("    glob pattern:", package)
        extended.extend(glob_result)
    return extended


def download_extract(
    url: str,
    dest: str | Path,
    dest_name: str,
    checksum: str = "",
) -> Path:
    """Download and extract a file from a URL."""
    name = Path(url).name
    with verbosity(2):
        print("Download URL:", url)
        # info("Download URL:", url)
    with tempfile.TemporaryDirectory() as tmp:
        target = Path(tmp) / name
        download_url(url, target)
        verify_checksum(target, checksum)
        dest_path = Path(dest) / dest_name
        unarchive(target, str(dest_path))
        with verbosity(2):
            print("Project path:", dest_path)
        with verbosity(3):
            sh(f"ls -l {dest_path}")
        return dest_path


def download_url(url: str, target: Path) -> None:
    try:
        with urlopen(url) as remote:  # noqa S310
            target.write_bytes(remote.read())
    except (HTTPError, OSError):
        warning(f"download_url() failed for {url}")
        raise


def verify_checksum(target: Path, checksum: str) -> None:
    """If a checksum is provided, verify that the Path has the same sha256
    hash.

    Abort the programm if hashes differ. Currently supporting only
    sha256.
    """
    if not checksum:
        return
    # Assuming checksum is a 64-long string verified by Nua-config
    file_hash = sha256(target.read_bytes()).hexdigest()
    if file_hash == checksum:
        with verbosity(2):
            show("Checksum of dowloaded file verified")
    else:
        raise Abort(
            f"Wrong checksum verification for file:\n{target}\n"
            f"Computed sha256 sum: {file_hash}\nExpected result:     {checksum}"
        )


def is_local_dir(project: str) -> bool:
    """Analysis of some ptoject string and guess wether local path or URL.

    (WIP)
    """
    parsed = urlparse(project)
    if parsed.scheme:
        # guess it is a download URL
        result = False
        abs_path_s = ""
    else:
        abs_path = Path(project).absolute().resolve()
        result = abs_path.is_dir()
        abs_path_s = str(abs_path)
    with verbosity(3):
        info(f"is_local_dir '{abs_path_s}'", result)
    return result


#
# String and replacement utils
#
def set_php_ini_keys(replacements: dict, path: str | Path):
    """Set keys in a php.ini file."""
    path = Path(path)
    content = path.read_text()
    for key, value in replacements.items():
        content = re.sub(rf"^;*\s*{key}\s*=", f"{key} = {value}", content, flags=re.M)
    path.write_text(content)


def replace_in(file_pattern: str, string_pattern: str, replacement: str):
    """Replace a string in a set of files, defined by a glob pattern."""
    with verbosity(2):
        show(f"replace_in() '{string_pattern}' by '{replacement}'")
    counter = 0
    for file_name in glob(file_pattern, recursive=True):
        path = Path(file_name)
        if not path.is_file():
            continue
        # assuming it's an utf8 world
        content = path.read_text(encoding="utf8")
        path.write_text(content.replace(string_pattern, replacement), encoding="utf8")
        counter += 1
    with verbosity(2):
        show(f"replace_in() done in {counter} files")


def string_in(file_pattern: str, string_pattern: str) -> list:
    """Return list of Path of files that contains the pattern str string."""
    hit_files = []
    # assuming it's an utf8 world
    upattern = string_pattern.encode("utf8")
    for file_name in glob(file_pattern, recursive=True):
        path = Path(file_name)
        if not path.is_file():
            continue
        with open(path, "rb", 0) as rfile, mmap.mmap(
            rfile.fileno(), 0, access=mmap.ACCESS_READ
        ) as mfile:
            if mfile.find(upattern) != -1:
                hit_files.append(path)
    return hit_files


def environ_replace_in(str_path: str | Path, env: dict | None = None):
    """Replace environment variables in a file."""
    path = Path(str_path)
    if not path.is_file():
        return
    orig_env = {}
    if env:
        orig_env = os.environ.copy()
        os.environ.update(env)
    try:
        # assuming it's an utf8 world
        content = path.read_text(encoding="utf8")
        path.write_text(os.path.expandvars(content), encoding="utf8")
    except OSError:
        raise
    finally:
        if env:
            os.environ.clear()
            os.environ.update(orig_env)


def append_bashrc(home: str | Path, content: str):
    """Append content to the bashrc file."""
    path = Path(home) / ".bashrc"
    if not path.is_file():
        raise FileNotFoundError(path)
    lines = f"# added by Nua script:\n{content}\n\n"
    with open(path, mode="a", encoding="utf8") as wfile:
        wfile.write(lines)


def snake_format(name: str) -> str:
    """Convert a string to snake_case format.
    >>> snake_format("my-project")
    'my_project'
    """
    return "_".join(word.lower() for word in name.replace("-", "_").split("_"))


def kebab_format(name: str) -> str:
    """Convert a string to kebab_case format.
    >>> kebab_format("my_project")
    'my-project'
    """
    return "-".join(word.lower() for word in name.replace("_", "-").split("-"))


def camel_format(name: str) -> str:
    """Convert a string to CamelCase format.
    >>> camel_format("my-project")
    'MyProject'
    """
    return "".join(word.title() for word in name.replace("-", "_").split("_"))


def _to_format_cases(
    formatter: Callable,
    data: dict[str, Any] | list,
    unchanged: Iterable,
    recurse: int = 999,
) -> None:
    """Converts all keys in a dict  or list to "formatter" format, with list of
    unchanged keys and recursion level, in place."""
    if isinstance(data, list):
        for item in data:
            if isinstance(item, dict):
                _to_format_cases_dict(formatter, item, unchanged, recurse - 1)
    elif isinstance(data, dict):
        _to_format_cases_dict(formatter, data, unchanged, recurse - 1)


def _to_format_cases_dict(
    formatter: Callable,
    data: dict[str, Any],
    unchanged: Iterable,
    recurse: int = 999,
) -> None:
    """Converts all keys in a dict to "formatter" format, with list of unchanged
    keys and recursion level, in place."""
    for key, value in list(data.items()):
        if key in unchanged:
            continue
        new_key = formatter(key)
        if new_key != key:
            data[new_key] = value
            del data[key]
        if recurse > 0 and isinstance(value, (dict, list)):
            _to_format_cases(formatter, value, unchanged, recurse - 1)


def to_snake_cases(
    data: dict[str, Any],
    recurse: int = 999,
    unchanged: Iterable | None = None,
) -> None:
    """Converts all keys in a dict to snake_case, recursion level,
    in place."""
    if unchanged is None:
        unchanged = []
    _to_format_cases(snake_format, data, unchanged, recurse)


def to_kebab_cases(
    data: dict[str, Any],
    recurse: int = 999,
    unchanged: Iterable | None = None,
) -> None:
    """Converts all keys in a dict to snake_case, recursion level,
    in place."""
    if unchanged is None:
        unchanged = []
    _to_format_cases(kebab_format, data, unchanged, recurse)


def copy_from_package(
    package: str,
    filename: str,
    destdir: Path,
    destname: str = "",
):
    """Copy a file from a package to a destination directory."""
    content = rso.files(package).joinpath(filename).read_text(encoding="utf8")
    if destname:
        target = destdir / destname
    else:
        target = destdir / filename
    target.write_text(content)



import os
from contextlib import contextmanager
from functools import cache
from pathlib import Path

STATE = {
    "verbose": 0,
    "colorize": True,
    "verbose_threshold": -1,
    "packages_updated": False,
}


def set_verbosity(value: int):
    assert isinstance(value, int)
    STATE["verbose"] = value


def verbosity_level() -> int:
    return STATE["verbose"]


def set_color(flag: bool):
    assert isinstance(flag, bool)
    STATE["colorize"] = flag
    if flag:
        os.environ.pop("NO_COLOR", None)
    else:
        os.environ["NO_COLOR"] = "1"


def use_color() -> bool:
    if "NO_COLOR" in os.environ or is_inside_container():
        return False
    return bool(STATE["colorize"])  # noqa


@contextmanager
def verbosity(level: int):
    previous = STATE["verbose_threshold"]
    STATE["verbose_threshold"] = level
    try:
        yield
    finally:
        STATE["verbose_threshold"] = previous


def check_verbosity() -> bool:
    return STATE["verbose"] >= STATE["verbose_threshold"]


@cache
def is_inside_container() -> bool:
    """Test if current execution environment is inside a container."""
    try:
        return Path("/nua/metadata/nua-config.json").is_file()
    except PermissionError:
        return False


def set_packages_updated(flag: bool) -> None:
    STATE["packages_updated"] = flag


def packages_updated() -> bool:
    return bool(STATE.get("packages_updated", False))



"""Nua scripting: print shortcuts.

All printing function should be encapsulated by a with verbosity(v) context for
future use in other message dispatching.


red bold: only for errors or serious warnings
    raise Abort()
    warning()  # got a Warning: prefix

red bold: warning lines
    red_line()

yellow bold: title of actions list
    title()
yellow bold + white: title of actions list
    title_info()

blue bold: when we’re about to start something important
    important()

green: when it’s done (i.e. for success messages)
    show()
green + cyan: information message with last part make visible
    info()

nocolor bold (default): : for normal immpotant messages (print)
    bold()

nocolor (default): for normal messages (print)
    vprint()  # with dependency on verbosity, so different from print
    vfprint() # and with stdout flush for rare special cases

magenta: for docker default log stream (without LF)
    print_stream()

gray: for debug messages (that are not supposed to be read, unless something goes
wrong or takes too much time)
    debug()

cyan: for emphasis in debug messages
    bold_debug()
"""
from typing import Any

from .console import (
    print_bold,
    print_bold_blue,
    print_bold_cyan,
    print_bold_red,
    print_bold_yellow,
    print_bold_yellow_white,
    print_cyan,
    print_green,
    print_green_cyan,
    print_magenta_no_lf,
)
from .tool.state import check_verbosity


class Abort(SystemExit):
    def __init__(
        self, msg: str = "unknown error", status: int = 1, explanation: str = ""
    ):
        if not msg:
            msg = "unknown error"
        _print_red_bold("Error", msg, explanation)
        super().__init__(status)


def warning(msg: str, explanation: str = ""):
    if check_verbosity():
        if not msg:
            msg = "unknown problem"
        _print_red_bold("Warning", msg, explanation)


def _print_red_bold(prefix: str, msg: str, explanation: str):
    if msg:
        msg = msg[0].upper() + msg[1:]
    print_bold_red(f"{prefix}: {msg}")
    if explanation:
        print_bold_red(f"    {explanation}")


def red_line(*args):
    if check_verbosity():
        print_bold_red(" ".join(str(x) for x in args))


def title(*args):
    if check_verbosity():
        print_bold_yellow(" ".join(str(x) for x in args))


def title_info(*args):
    if check_verbosity():
        print_bold_yellow_white(" ".join(str(x) for x in args))


def important(*args):
    if check_verbosity():
        print_bold_blue(" ".join(str(x) for x in args))


def show(*args: Any):
    if check_verbosity():
        print_green(" ".join(str(x) for x in args))


def info(*args: Any):
    if check_verbosity():
        print_green_cyan(" ".join(str(x) for x in args))


def bold(*args):
    if check_verbosity():
        print_bold(" ".join(str(x) for x in args))


def vprint(*args: Any):
    if check_verbosity():
        print(*args)


def vfprint(*args: Any):
    if check_verbosity():
        print(*args, end="", flush=True)


def print_stream(message: str):
    """Print Docker log stream already containing a LF."""
    if check_verbosity():
        print_magenta_no_lf(message)


def debug(*args: Any):
    if check_verbosity():
        print_cyan(" ".join(str(x) for x in args))


def bold_debug(*args: Any):
    if check_verbosity():
        print_bold_cyan(" ".join(str(x) for x in args))



from functools import wraps

from .tool.color_str import ColorStr
from .tool.state import use_color


def if_color(func):
    @wraps(func)
    def wrapper(*args):
        if use_color():
            return func(*args)
        else:
            return print(*args)

    return wrapper


@if_color
def print_green(msg: str):
    print(ColorStr.green(msg))


@if_color
def print_gray(msg: str):
    print(ColorStr.gray(msg))


@if_color
def print_blue(msg: str):
    print(ColorStr.blue(msg))


@if_color
def print_bold_blue(msg: str):
    print(ColorStr.blue_bold(msg))


@if_color
def print_magenta(msg: str):
    print(ColorStr.magenta(msg))


@if_color
def print_red(msg: str):
    print(ColorStr.red(msg))


@if_color
def print_bold_red(msg: str):
    print(ColorStr.red_bold(msg))


@if_color
def print_bold(msg: str):
    print(ColorStr.colorize(msg, bold=True))


@if_color
def print_bold_yellow(msg: str):
    print(ColorStr.yellow_bold(msg))


@if_color
def print_cyan(msg: str):
    print(ColorStr.cyan(msg))


@if_color
def print_bold_cyan(msg: str):
    print(ColorStr.cyan_bold(msg))


@if_color
def print_bold_yellow_white(msg: str):
    parts = msg.rsplit(" ", 1)
    if len(parts) == 2:
        print(ColorStr.yellow_bold(parts[0]), ColorStr.white_bold(parts[1]))
    else:
        print(ColorStr.yellow_bold(msg))


@if_color
def print_green_cyan(msg: str):
    parts = msg.rsplit(" ", 1)
    if len(parts) == 2:
        print(ColorStr.green(parts[0]), ColorStr.cyan(parts[1]))
    else:
        print(ColorStr.green(msg))


@if_color
def print_magenta_no_lf(msg: str):
    """Print specialized for lines including a LF (Docker build log)."""
    print(ColorStr.magenta(msg), end="")



"""Class to manage the "nua-config.toml" file."""
import json
import os
from copy import deepcopy
from functools import cached_property
from pathlib import Path
from typing import Any

import tomli
import yaml
from dictdiffer import diff

from .actions import download_extract, to_kebab_cases, to_snake_cases
from .constants import NUA_CONFIG_STEM, nua_config_names
from .exceptions import NuaConfigError
from .normalization import normalize_env_values, normalize_ports, ports_as_list
from .nua_config_format import NuaConfigFormat
from .nua_tag import nua_tag_string
from .panic import red_line, vprint
from .shell import chown_r

# blocks added (empty) if not present in orig file:
COMPLETE_BLOCKS = ["build", "run", "env", "docker"]


def hyphen_get(data: dict, key: str, default: Any = None) -> Any:
    """Return value from dict for key either hyphen "-" or underscore "_" in
    it, priority to "-"."""
    if (hypkey := key.replace("_", "-")) in data:
        result = data.get(hypkey)
    elif (underkey := key.replace("-", "_")) in data:
        result = data.get(underkey)
    else:
        result = default
    return result


def force_list(content: Any) -> list[Any]:
    """Return always a list, if a single string is provided, wrap it into
    list.
    >>> force_list("foo")
    ['foo']
    >>> force_list(["foo", "bar"])
    ['foo', 'bar']
    >>> force_list([])
    []
    >>> force_list("")
    []
    """
    if content is None or content == "":
        return []
    match content:
        case list():
            return content
        case _:
            return [content]


class NuaConfig:
    """Class to manage the "nua-config.toml" file.

    The config file can be any of nua-config.[toml|json|yaml|yml]
    """

    path: Path
    root_dir: Path
    nua_dir_exists: bool
    _data: dict[str, Any]

    def __init__(self, path: str | Path | None = None):
        if not path:
            path = "."
        self._find_root_dir(path)
        self._find_config_file()
        self._loads_config()
        self._check_checksum_format()
        self._normalize_env_values()
        self._normalize_ports()

    def __repr__(self):
        return f"NuaConfig(path={self.path}, id={id(self)})"

    def _find_root_dir(self, path: Path | str) -> None:
        """Find project folder.

        Analyses structure of parameters:

        Below the expected structure of folders. 'nua' folder and subfolders are
        optional, the 'nua-config' file can be at 3 positions: 'project', 'nua'
        subfolder, or 'metadata' subfolder.

        The path parameter can be:
            - the 'nua-config' file
            - the path of 'nua' folder, 'project' folder or folder containing the
              'nua-config' file

        We try to:
            - find the 'nua-file'
            - find the 'project' path
            - find the 'nua' path (if exists)

        The 'nua-config' file used is the one of higher level found.

        /project
            ...project code source...
            nua-config
            /nua
                Dockerfile
                build.py
                ...
                nua-config
                /metadata
                    nua-config
                /build
                   ...files to copy...
                /app
                   ...files to copy...
        """
        path = Path(path).resolve()
        if path.is_dir():
            folder = path
        else:
            folder = path.parent
        if folder.name == "nua":
            self.root_dir = folder.parent
        elif folder.name == "metadata" and folder.parent.name == "nua":
            self.root_dir = folder.parent.parent
        else:
            self.root_dir = folder
        self.nua_dir_exists = (self.root_dir / "nua").is_dir()

    def _find_config_file(self):
        """Find nua-config file."""
        for folder in (
            self.root_dir,
            self.root_dir / "nua",
            self.root_dir / "nua" / "metadata",
        ):
            if not folder.is_dir():
                continue
            for name in nua_config_names():
                test_path = folder / name
                if test_path.is_file():
                    self.path = test_path
                    return

        raise NuaConfigError(
            f"Nua config file not found in '{self.root_dir}' and sub folders"
        )

    def _read_config_data(self) -> dict[str, Any]:
        if self.path.suffix == ".toml":
            return tomli.loads(self.path.read_text(encoding="utf8"))
        elif self.path.suffix in {".json", ".yaml", ".yml"}:
            return yaml.safe_load(self.path.read_text(encoding="utf8"))
        else:
            raise NuaConfigError(f"Unknown file extension for '{self.path}'")

    def _loads_config(self) -> None:
        unchanged = {"docker", "env"}
        recurse = 4
        read_data = self._read_config_data()
        data = deepcopy(read_data)
        to_snake_cases(data, recurse=recurse, unchanged=unchanged)
        # store internally as a dict() after pydantic validation:
        data = NuaConfigFormat(**data).dict()
        to_kebab_cases(data, recurse=recurse, unchanged=unchanged)
        self._data = data
        self._show_config_differences(read_data)

    def _show_config_differences(self, data: dict[str, Any]) -> None:
        """Print the difference between original data and actual loaded config.

        This shows the diff between the original TOML and the parsed config once
        it has been converted to a Pydantic model.
        Pydantic doesn't check (AFAIK) for extra (just for missing keys or values
        of the wrong type), so this is a way to check for typos in the keys, etc.
        """
        header_displayed = False
        for key, section, changes in diff(data, self._data):
            if len(changes) == 2 and not isinstance(changes[0], tuple):
                change_list = [changes]
            else:
                change_list = changes
            for change_key, value in change_list:
                if value is None:
                    continue
                if not header_displayed:
                    red_line("NuaConfig data changes:")
                    header_displayed = True
                vprint(f"    {key}: {section}.{change_key}: {repr(value)}")

    def as_dict(self) -> dict[str, Any]:
        return self._data

    def dump_json(self, folder: Path | str) -> None:
        dest = Path(folder) / f"{NUA_CONFIG_STEM}.json"
        dest.write_text(
            json.dumps(self._data, sort_keys=False, ensure_ascii=False, indent=4),
            encoding="utf8",
        )

    def _check_checksum_format(self):
        checksum = self.src_checksum
        if not checksum:
            return
        if checksum.startswith("sha256:"):
            checksum = checksum[7:]
        if len(checksum) != 64:
            raise NuaConfigError(
                f"Wrong src-checksum content (expecting 64 length sha256): {checksum}"
            )

    def _normalize_env_values(self):
        self._data["env"] = normalize_env_values(self.env)

    def _normalize_ports(self):
        normalized_ports = normalize_ports(ports_as_list(self.ports))
        self._data["port"] = {}
        for port in normalized_ports:
            name = port["name"]
            del port["name"]
            self._data["port"][name] = port

    def __getitem__(self, key: str) -> Any:
        """will return {} if key not found, assuming some parts are not
        mandatory and first level element are usually dict."""
        return self._data.get(key) or {}

    # metadata ######################################################

    @property
    def metadata(self) -> dict:
        return self["metadata"]

    @cached_property
    def metadata_rendered(self) -> dict:
        """Return the metadata and build sections with rendered
        f-string values."""
        mixed = deepcopy(self.metadata)
        mixed.update(deepcopy(self.build))
        data = {}
        for key, val in deepcopy(mixed).items():
            if isinstance(val, str):
                data[key] = val.format(**mixed)
            else:
                data[key] = val
        return data

    @property
    def version(self) -> str:
        """Version of package source."""
        return self.metadata.get("version", "")

    @property
    def app_id(self) -> str:
        return self.metadata.get("id", "")

    @property
    def title(self) -> str:
        return self.metadata.get("title", "")

    @property
    def nua_tag(self) -> str:
        return nua_tag_string(self.metadata)

    # build #########################################################

    @property
    def build(self) -> dict:
        return self._data.get("build") or {}

    @cached_property
    def src_url(self) -> str:
        if base := hyphen_get(self.build, "src-url"):
            return base.format(**self.metadata_rendered)
        return ""  #

    @cached_property
    def git_url(self) -> str:
        if base := hyphen_get(self.build, "git-url"):
            return base.format(**self.metadata_rendered)
        return ""

    @cached_property
    def git_branch(self) -> str:
        if base := hyphen_get(self.build, "git-branch"):
            return base.format(**self.metadata_rendered)
        return "main"

    @cached_property
    def wrap_image(self) -> str:
        """Optional container image to be used as base for 'wrap' strategy.

        If the 'base-image' metadata is defined, the build strategy is to add
        the '/nua/metadata/nua-config.json' file on the declared image,
        when build method is 'wrap'.
        """
        if base := hyphen_get(self.build, "base-image"):
            return base.format(**self.metadata_rendered)
        return ""

    @property
    def src_checksum(self) -> str:
        """Return checksum associated to 'src-url' or null string."""
        if checksum := hyphen_get(self.build, "src-checksum"):
            return checksum.strip().lower()
        return ""  #

    @cached_property
    def builder(self) -> str | dict[str, str] | list[dict]:
        if base := self.build.get("builder", ""):
            if isinstance(base, str):
                return base.format(**self.metadata_rendered)
        return ""

    @cached_property
    def project(self) -> str:
        """The project URL to build with autodetection."""
        if base := self.build.get("project"):
            return base.format(**self.metadata_rendered)
        return ""

    @property
    def manifest(self) -> list:
        return force_list(self.build.get("manifest") or "")

    @property
    def meta_packages(self) -> list:
        return force_list(hyphen_get(self.build, "meta-packages") or "")

    @property
    def build_packages(self) -> list:
        return force_list(self.build.get("packages") or "")

    @cached_property
    def build_command(self) -> list:
        """Return the list of build commands, each cmd rendered with
        metadata."""
        commands = []
        for cmd in force_list(self.build.get("build", [])):
            if isinstance(cmd, str):
                commands.append(cmd.format(**self.metadata_rendered))
            else:
                commands.append(cmd)
        return commands

    @property
    def pip_install(self) -> list:
        return force_list(hyphen_get(self.build, "pip-install", []))

    @property
    def build_method(self) -> str:
        """Build method (or default build method).

        Can be empty string if not defined (then autodetection from
        metadata).
        """
        default_method = hyphen_get(self.build, "default-method", "")
        return self.build.get("method", default_method)

    # run ###########################################################

    @property
    def run(self) -> dict:
        return self._data.get("run") or {}

    @property
    def packages(self) -> list:
        "Return list of run packages (packages remanent after build)."
        return force_list(self.run.get("packages", []))

    @property
    def start_command(self) -> list:
        return force_list(hyphen_get(self.run, "start", []))

    # env ###########################################################

    @property
    def env(self) -> dict:
        return self._data.get("env") or {}

    # docker env ####################################################

    @property
    def docker(self) -> dict:
        return self._data.get("docker") or {}

    # volumes declaration ###########################################

    @property
    def volumes(self) -> list:
        """The list of declared volumes."""
        return self._data.get("volume") or []

    # providers #####################################################

    @property
    def providers(self) -> list:
        """The list of providers (tag 'provider')."""
        return self._data.get("provider") or []

    # ports #########################################################

    @property
    def ports(self) -> dict[str, dict]:
        return self._data.get("port") or {}

    # actions #######################################################

    def fetch_source(self, name: str = "") -> Path:
        """Download, check and extract source URL, returning the source Path.

        Set ownership to 'nua' user if launched by 'root'.
        """
        if not name:
            name = self.app_id
            if name.startswith("nua-"):
                name = name[4:]
        path = download_extract(self.src_url, "/nua/build", name, self.src_checksum)
        if os.getuid() == 0:
            chown_r(path, "nua")
        return path



from ..backports import chdir
from ..shell import sh
from .apt import apt_remove_lists, install_package_list, purge_package_list
from .misc import compile_openssl_1_1


def _build_ruby_install() -> str:
    """Download and install 'ruby-install' program.

    Returns:
        ruby-install path
    """
    ri_vers = "0.9.1"

    with chdir("/tmp"):  # noqa s108
        cmd = (
            f"wget -O ruby-install-{ri_vers}.tar.gz "
            f"https://github.com/postmodern/ruby-install/archive/v{ri_vers}.tar.gz"
        )
        sh(cmd)
        cmd = f"tar -xzvf ruby-install-{ri_vers}.tar.gz"
        sh(cmd)
        with chdir(f"ruby-install-{ri_vers}"):
            cmd = "make install"
            sh(cmd)

    cmd = f"rm -fr /tmp/ruby-install-{ri_vers}*"
    sh(cmd)
    return "/usr/local/bin/ruby-install"


def install_ruby(
    version: str = "3.2.1",
    keep_lists: bool = False,
) -> None:
    """Installation of Ruby via 'ruby-install'.

    Exec as root.

    Args:
        version: Ruby version to install.
        keep_lists: Do not erase apt sources lists.
    """
    install_package_list(
        "wget pkg-config build-essential libpq-dev",
        keep_lists=True,
        clean=False,
    )
    purge_package_list("ruby ruby-dev ri")
    options = "--disable-install-doc"

    if version.startswith("2.") or version.startswith("3.0"):
        ssl = compile_openssl_1_1()
        options = f"--disable-install-doc --with-openssl-dir={ssl}"

    path = _build_ruby_install()
    cmd = f"{path} --system --cleanup -j4 {version} -- {options}"
    sh(cmd)

    if not keep_lists:
        apt_remove_lists()



from copy import deepcopy
from typing import Any

from .exceptions import NuaConfigError


def normalize_env_values(env: dict[str, str | int | float | list]) -> dict[str, Any]:
    def normalize_env_leaf(value: Any) -> str:
        if isinstance(value, str):
            return value
        if isinstance(value, (int, float, list)):
            return str(value)
        raise NuaConfigError(f"ENV value has wrong type: '{value}'")

    validated: dict[str, str | dict] = {}
    for key, value in env.items():
        if isinstance(value, dict):
            validated[key] = {k: normalize_env_leaf(v) for k, v in value.items()}
        else:
            validated[key] = normalize_env_leaf(value)

    return deepcopy(validated)


def normalize_keys(provider: dict[str, Any]) -> None:
    for key, value in provider.items():
        if "_" in key:
            new_key = key.replace("_", "-")
            provider[new_key] = value
            del provider[key]


def ports_as_list(ports_dict: dict) -> list:
    keys = list(ports_dict.keys())
    for key in keys:
        ports_dict[key]["name"] = key
    return list(ports_dict.values())


def normalize_ports(port_list: list[dict]) -> list[dict]:
    for port in port_list:
        _normalize_port(port)
    return port_list


def _normalize_port(port: dict[str, Any]) -> None:
    try:
        # _normalize_port_item_web(port)
        _normalize_port_item_container(port)
        _normalize_port_item_host(port)
        _normalize_port_item_proxy(port)
        _set_web_flag(port)
        _normalize_port_item_protocol(port)
        _normalize_port_item_ssl(port)
    except NuaConfigError:
        print("--- Error in ports config: ---")
        print(port)
        raise


# def _normalize_port_item_web(port: dict[str, Any]) -> None:
#     name = port["name"].lower()
#     web = port["web"]
#     # special case if  name of port is web: force web to True, refuse False
#     if name == "web":
#         if web is False:
#             raise NuaConfigError(
#                 "port['web'] can not be False for the port named 'web'"
#             )
#         web = True
#     if web is None:
#         # Defaulting to False
#         # The rule is : for any port with name different of "web", it's
#         # mandatory to set the web key to True, or the port is not a
#         # proxyied by nginx (for security resaon)
#         web = False
#     port["web"] = web


def _normalize_port_item_container(port: dict[str, Any]) -> None:
    pass


def _normalize_port_item_host(port: dict[str, Any]) -> None:
    # if port["web"]:
    #     # allow either None or integer host port for web traffic
    #     return
    host = port["host"]
    if isinstance(host, dict):
        return
    # if not port["host"], requirements: either some web proxy or name
    # of port is web for # automatic proxy
    if not host and not port["proxy"] and port["name"].lower() != "web":
        raise NuaConfigError("port['host'] is required for non-web traffic")


def _normalize_port_item_proxy(port: dict[str, Any]) -> None:
    """A web port must have a proxy values, except for the "web" named port,
    which is automatic (so either proxied to 80 or 443).

    For a non web port (ie. mail on port 25), the proxy is not used currently
    (unless some firewall is configured in the future). But the 'host' port is
    required.
    So a port without proxy must have a host port.
    """
    proxy = port["proxy"]
    if isinstance(proxy, dict):
        return
    # if not port["web"]:
    #     # allow proxy defined or not
    #     return
    if not proxy and not port["host"] and port["name"].lower() != "web":
        raise NuaConfigError(
            (
                "Only port.web can have an automatic 'proxy' value, other web "
                "published ports muste provide a proxy number"
            )
        )


def _set_web_flag(port: dict[str, Any]) -> None:
    """Set the key 'web' to True if the port is meaned to be proxyied by nginx."""
    if port["name"].lower() == "web" or port["proxy"]:
        port["web"] = True
    else:
        port["web"] = False


def _normalize_port_item_protocol(port: dict[str, Any]) -> None:
    protocol = port["protocol"]
    if protocol is None:
        protocol = "tcp"
    protocol = str(protocol).lower().strip()
    if protocol not in {"tcp", "udp"}:
        raise NuaConfigError(
            "port['protocol'] must be 'tcp' or 'udp' (default is 'tcp')"
        )
    port["protocol"] = protocol


def _normalize_port_item_ssl(port: dict[str, Any]) -> None:
    ssl = port["ssl"]
    if ssl is None:
        ssl = True
    port["ssl"] = ssl



import os
from collections.abc import Iterable
from contextlib import contextmanager

from ..panic import show, warning
from ..shell import sh
from ..tool.state import packages_updated, set_packages_updated, verbosity
from .constants import LONG_TIMEOUT, SHORT_TIMEOUT


@contextmanager
def install_build_packages(
    packages: list | str,
    keep_lists: bool = False,
    installed: Iterable[str] | None = None,
):
    """Install packages needed for building a project."""
    success = True

    if installed:
        already = set(installed)
    else:
        already = set()

    if isinstance(packages, str):
        packages = packages.strip().split()

    needed = [package for package in packages if package not in already]
    if needed:
        with verbosity(2):
            show("Install temporary build packages")
        _install_packages(needed)

    try:
        yield
    except SystemExit:
        success = False
        raise
    finally:
        if success:
            if needed:
                with verbosity(2):
                    show("Remove temporary build packages")
                _purge_packages(needed)
            apt_final_clean()
            if not keep_lists:
                apt_remove_lists()


def apt_remove_lists():
    environ = os.environ.copy()
    sh("rm -rf /var/lib/apt/lists/*", env=environ, timeout=SHORT_TIMEOUT)
    set_packages_updated(False)


def apt_update():
    environ = os.environ.copy()
    environ["DEBIAN_FRONTEND"] = "noninteractive"
    cmd = "apt-get update --fix-missing"
    sh(cmd, env=environ, timeout=LONG_TIMEOUT, show_cmd=False)


def apt_final_clean():
    environ = os.environ.copy()
    environ["DEBIAN_FRONTEND"] = "noninteractive"
    cmd = "apt-get autoremove -y; apt-get clean"
    sh(cmd, env=environ, timeout=SHORT_TIMEOUT)


def _install_packages(packages: list):
    if not packages:
        warning("install_package(): nothing to install")
        return

    environ = os.environ.copy()
    environ["DEBIAN_FRONTEND"] = "noninteractive"
    if not packages_updated():
        cmd = "apt-get update --fix-missing; "
    else:
        cmd = ""
    cmd += f"apt-get install --no-install-recommends -y {' '.join(packages)}"
    sh(cmd, env=environ, timeout=LONG_TIMEOUT)
    set_packages_updated(True)


def _purge_packages(packages: list):
    if not packages:
        return

    print(f"Purge packages: {' '.join(packages)}")
    environ = os.environ.copy()
    environ["DEBIAN_FRONTEND"] = "noninteractive"
    for package in packages:
        cmd = f"apt-get purge -y {package} 2>/dev/null || true"
        sh(cmd, env=environ, timeout=SHORT_TIMEOUT, show_cmd=False)


def install_packages(packages: list, keep_lists: bool = False):
    if packages:
        with verbosity(2):
            show("install packages")
        install_package_list(packages, keep_lists=keep_lists)


def install_package_list(
    packages: list | str,
    clean: bool = True,
    keep_lists: bool = False,
):
    if isinstance(packages, str):
        packages = packages.strip().split()
    _install_packages(packages)
    if clean:
        apt_final_clean()
    if not keep_lists:
        apt_remove_lists()


def purge_package_list(packages: list | str):
    if isinstance(packages, str):
        packages = packages.strip().split()
    _purge_packages(packages)
    apt_final_clean()


@contextmanager
def tmp_install_package_list(
    packages: list | str,
    keep_lists: bool = True,
):
    if isinstance(packages, str):
        packages = packages.strip().split()
    _install_packages(packages)
    try:
        yield
    finally:
        _purge_packages(packages)
        apt_final_clean()
        if not keep_lists:
            apt_remove_lists()


def installed_packages() -> set[str]:
    cmd = "apt list --installed"
    result = sh(cmd, timeout=SHORT_TIMEOUT, capture_output=True, show_cmd=False)
    return {name.split("/")[0] for name in str(result).splitlines()}



"""Class to manage NuaConfig format."""
from typing import Any

from pydantic import BaseModel, NonNegativeInt, PositiveInt


class Metadata(BaseModel):
    id: str
    version: str
    title: str
    author: str
    description: str | None
    tagline: str | None
    website: str | None
    tags: list[str] | None
    profile: str | list[str] | None
    release: PositiveInt | None


class Build(BaseModel):
    license: str = "Proprietary"
    src_url: str | None
    src_checksum: str | None
    git_url: str | None
    git_branch: str | None
    base_image: str | None
    method: str | None
    builder: str | dict | None
    builders: list[str] | list[dict] | None
    packages: str | list[str] | None
    meta_packages: str | list[str] | None
    build: str | list[str] | None
    test: str | list[str] | None
    before_build: str | list[str] | None
    pip_install: str | list[str] | None
    project: str | None

    # Obsolete, relaced by 'build':
    # build_command: str | list[str] | None  # add


class Run(BaseModel):
    packages: str | list[str] | None
    before_run: str | list[str] | None
    start: str | list[str] | None
    post_run: str | list[str] | None
    post_run_status: str | None

    # Should the '[env]' section be here or at the root ?
    # env: dict[str, Any] | None


class Port(BaseModel):
    container: PositiveInt | dict
    host: PositiveInt | dict | None
    proxy: PositiveInt | dict | None
    protocol: str | None
    ssl: bool | None


class Healthcheck(BaseModel):
    command: str
    interval: PositiveInt
    start_period: PositiveInt | None
    timeout: PositiveInt | None
    retries: NonNegativeInt | None


class Backup(BaseModel):
    method: str
    frequency: str | None
    options: dict[str, Any] | None


class Volume(BaseModel):
    name: str | None
    type: str | None
    target: str
    driver: str | None
    options: dict | None
    backup: Backup | None


class Provider(BaseModel):
    name: str
    module_name: str | None
    module_version: str | None
    module_versions: list[dict] | None
    type: str | None
    network: bool | None
    connector: str | list[str] | None
    build: Build | None
    run: Run | None
    env: dict[str, Any] | None
    port: dict[str, Port] | None
    volume: list[Volume] | None
    docker: dict | None
    backup: Backup | None


class NuaConfigFormat(BaseModel):
    metadata: Metadata
    build: Build | None
    run: Run | None
    env: dict[str, Any] | None
    provider: list[Provider] | None  #
    volume: list[Volume] | None  #
    port: dict[str, Port] | None
    docker: dict | None
    backup: Backup | None
    # Should it be in `run` ?
    healthcheck: Healthcheck | None  #



"""Return text string colorized.

from color_str import ColorStr as C print(C.green_bright('Take') +
C.blue(' it ', underline=True)       + C.yellow_bold('easy',
bgcolor='red'))
"""
from functools import cache, partialmethod


class ColorStr:
    COL = {
        "black": "\u001b[30",
        "red": "\u001b[31",
        "green": "\u001b[32",
        "yellow": "\u001b[33",
        "blue": "\u001b[34",
        "magenta": "\u001b[35",
        "cyan": "\u001b[36",
        "white": "\u001b[37",
        "reset": "\u001b[0m",
        "bright": ";1",
        "suffix": "m",
        "bold": "\u001b[1m",
        "underline": "\u001b[4m",
        "reversed": "\u001b[7m",
        "bg_black": "\u001b[40",
        "bg_red": "\u001b[41",
        "bg_green": "\u001b[42",
        "bg_yellow": "\u001b[43",
        "bg_blue": "\u001b[44",
        "bg_magenta": "\u001b[45",
        "bg_cyan": "\u001b[46",
        "bg_white": "\u001b[47",
    }

    @classmethod
    def _color(cls, cmd: list, color: str, bright: bool):
        cmd.append(cls.COL[color])
        if bright:
            cmd.append(cls.COL["bright"])
        cmd.append(cls.COL["suffix"])

    @classmethod
    @cache
    def _command(cls, color: str = "", bgcolor: str = "", **kwargs) -> str:
        cmd = []
        for key in ("bold", "underline", "reversed"):
            if key in kwargs:
                cmd.append(cls.COL[key])
        if color:
            cls._color(cmd, color, bool(kwargs.get("bright")))
        if bgcolor:
            cls._color(cmd, f"bg_{bgcolor}", bool(kwargs.get("bright")))
        return "".join(cmd)

    @classmethod
    def colorize(
        cls,
        txt: str,
        color: str = "",
        bgcolor: str = "",
        **kwargs,
    ) -> str:
        """Return text string colorized.

        kwargs are bools for keys "bright", "bold", "underline" and
        "reversed".
        """
        cmd = cls._command(color=color, bgcolor=bgcolor, **kwargs)
        return f"{cmd}{txt}{cls.COL['reset']}"

    black = partialmethod(colorize, color="black")
    gray = partialmethod(colorize, color="black", bright=True)
    red = partialmethod(colorize, color="red")
    green = partialmethod(colorize, color="green")
    yellow = partialmethod(colorize, color="yellow")
    blue = partialmethod(colorize, color="blue")
    magenta = partialmethod(colorize, color="magenta")
    cyan = partialmethod(colorize, color="cyan")
    white = partialmethod(colorize, color="white")
    black_bold = partialmethod(colorize, color="black", bold=True)
    red_bold = partialmethod(colorize, color="red", bold=True)
    green_bold = partialmethod(colorize, color="green", bold=True)
    yellow_bold = partialmethod(colorize, color="yellow", bold=True)
    blue_bold = partialmethod(colorize, color="blue", bold=True)
    magenta_bold = partialmethod(colorize, color="magenta", bold=True)
    cyan_bold = partialmethod(colorize, color="cyan", bold=True)
    white_bold = partialmethod(colorize, color="white", bold=True)
    black_bright = partialmethod(colorize, color="black", bright=True)
    red_bright = partialmethod(colorize, color="red", bright=True)
    green_bright = partialmethod(colorize, color="green", bright=True)
    yellow_bright = partialmethod(colorize, color="yellow", bright=True)
    blue_bright = partialmethod(colorize, color="blue", bright=True)
    magenta_bright = partialmethod(colorize, color="magenta", bright=True)
    cyan_bright = partialmethod(colorize, color="cyan", bright=True)
    white_bright = partialmethod(colorize, color="white", bright=True)



"""Scripting shell utils for Nua scripts."""
import pwd
import shutil
from pathlib import Path
from subprocess import CompletedProcess, run  # noqa: S404

from .panic import Abort, show


def cat(filename: str | Path):
    with open(filename, encoding="utf8") as fd:
        print(fd.read())


def chown_r(path: str | Path, user: str, group: str | None = None):
    """Apply recursively chown with str arguments.

    Similar to `chown -R`.

    example:
        chown_r(document_root, "www-data", "www-data")
    """
    root = Path(path).absolute()
    # broken on symlinks ?
    # for item in root.rglob("*"):
    #     print(item, item.is_file())
    #     shutil.chown(item, user, group or user)
    # shutil.chown(root, user, group or user)
    sh(f"chown -R {user}:{group or user} {root}", show_cmd=False)


def chmod_r(path: str | Path, file_mode: int, dir_mode: int | None = None):
    """Apply recursively chmod with int arguments.

    Similar to `chmod -R` but with int arguments (usually expressed as octal numbers).

    example:
        chmod_r(document_root, 0o644, 0o755)
    """
    root = Path(path)
    if dir_mode is None:
        dir_mode = file_mode
    for item in root.rglob("*"):
        if item.is_dir():
            item.chmod(dir_mode)
        else:
            item.chmod(file_mode)
    if root.is_dir():
        root.chmod(dir_mode)
    else:
        root.chmod(file_mode)


# XXX: not used
def _dir_chmod_r(root: Path, file_mode: int, dir_mode: int):
    for subpath in root.rglob(""):
        if subpath.is_dir():
            subpath.chmod(dir_mode)
        else:
            subpath.chmod(file_mode)


def echo(text: str, filename: str | Path) -> None:
    Path(filename).write_text(text + "\n", encoding="utf8")


def mkdir_p(path: str | Path):
    Path(path).mkdir(parents=True, exist_ok=True)


def rm_fr(path: str | Path) -> bool:
    """Alias for rm_rf."""
    return rm_rf(path)


def rm_rf(path: str | Path) -> bool:
    """Wrapper for shutil.rmtree()."""
    if Path(path).exists():
        shutil.rmtree(path)
        return True
    return False


def _base_sh(
    cmd: str,
    timeout: int | None,
    env: dict | None,
    capture_output: bool,
    cwd: str | Path | None,
) -> CompletedProcess:
    if capture_output:
        return run(
            cmd,
            shell=True,  # noqa: S602
            executable="/bin/bash",
            timeout=timeout,
            env=env,
            capture_output=True,
            encoding="utf8",
            text=True,
            cwd=cwd,
        )
    else:
        return run(
            cmd,
            shell=True,  # noqa: S602
            executable="/bin/bash",
            timeout=timeout,
            env=env,
            cwd=cwd,
        )


def sh(
    cmd: str,
    timeout: int | None = 600,
    env: dict | None = None,
    show_cmd: bool = True,
    capture_output: bool = False,
    cwd: str | Path | None = None,
) -> str | bytes:
    # XXX: can/should it really return bytes?
    """Run a shell command."""
    if show_cmd:
        show(cmd)

    try:
        # subprocess call with shell=True identified, security issue:
        # We do want to mimic current shell action, including all environment
        completed = _base_sh(cmd, timeout, env, capture_output, cwd)
        status = completed.returncode
        if status < 0:
            msg = (
                f"Child was terminated by signal {-status},\n"
                f"shell command was: '{cmd}'"
            )
            raise Abort(msg, status)
        elif status > 0:
            msg = (
                f"Something went wrong (exit code: {status}), \n"
                f"shell command was: '{cmd}'\n"
                f"{completed.stdout}\n"
                f"{completed.stderr}"
            )
            raise Abort(msg, status)
    except OSError as e:
        raise Abort(f"Execution failed: {e}\nshell command was: '{cmd}'")

    if capture_output:
        return completed.stdout
    else:
        return ""


def user_exists(user: str) -> bool:
    try:
        pwd.getpwnam(user)
        return True
    except KeyError:
        return False



"""Low-level utilities for both NUA build and NUA agent."""

# Version of module available at first level, __version__
# computed from pyproject data in version.py module
from .version import __version__  # noqa: F401



def elapsed(delta: float) -> str:
    """Readable time string from duration in seconds.
    >>> elapsed(123.456)
    '2min 3s'
    >>> elapsed(23.456)
    '23.46s'
    """

    seconds = int(delta)
    cent = int(round((delta - seconds) * 100))
    days, seconds = divmod(seconds, 86400)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    if days > 0:
        message = f"{days}d {hours}h {minutes}min"
    elif hours > 0:
        message = f"{hours}h {minutes}min {seconds}s"
    elif minutes > 0:
        message = f"{minutes}min {seconds}s"
    else:
        message = f"{seconds}.{cent:02d}s"
    return message



import abc
import tarfile
from pathlib import Path
from zipfile import ZipFile


class Unarchiver(abc.ABC):
    """ABC for unarchivers."""

    accepted_suffixes: list[str] = []

    @classmethod
    def accept(cls, src: str | Path) -> bool:
        src = str(src)
        for suffix in cls.accepted_suffixes:
            if src.endswith(suffix):
                return True
        return False

    @classmethod
    @abc.abstractmethod
    def extract(cls, src: str, dest_dir: str) -> None:
        raise NotImplementedError()


class TarUnarchiver(Unarchiver):
    accepted_suffixes = [".tar", ".tar.gz", ".tar.bz2", ".tar.xz", ".tgz"]

    @classmethod
    def extract(cls, src: str, dest_dir: str) -> None:
        def get_members(tar, root):
            for member in tar.getmembers():
                path = Path(member.path)
                member.path = path.relative_to(root)
                yield member

        with tarfile.open(src) as tar:
            root = Path(tar.getmembers()[0].path)
            tar.extractall(dest_dir, members=get_members(tar, root))  # noqa s202


class ZipUnarchiver(Unarchiver):
    accepted_suffixes = [".zip"]

    @classmethod
    def extract(cls, src: str, dest_dir: str) -> None:
        dest_path = Path(dest_dir)
        with ZipFile(src) as archive:
            arch_root = Path(archive.infolist()[0].filename)
            for member in archive.infolist():
                path = Path(member.filename)
                rela_path = path.relative_to(arch_root)
                target_path = dest_path.joinpath(rela_path)
                if member.is_dir():
                    target_path.mkdir(parents=True, exist_ok=True)
                    continue
                with archive.open(member) as content:
                    target_path.write_bytes(content.read())


def unarchive(src: str | Path, dest_dir: str) -> None:
    unarchivers = [TarUnarchiver, ZipUnarchiver]
    for unarchiver in unarchivers:
        if unarchiver.accept(src):
            unarchiver.extract(str(src), dest_dir)
            return
    raise ValueError(f"Unknown archive format for '{src}'")



"""Nua scripting: exec commands."""
import multiprocessing as mp
import os
import pwd
import shlex
from subprocess import DEVNULL, run  # noqa S404

NUA = "nua"


def sudo_cmd_as_user(  # noqa CFQ002
    cmd: str | list,
    user: str,
    *,
    cwd: str | None = None,
    timeout: int | None = None,
    env: dict | None = None,
    show_cmd: bool = True,
    set_home=True,
):
    if isinstance(cmd, list):
        cmd = " ".join(cmd)
    sudo_cmd = f"sudo -nu {user} {cmd}"
    if show_cmd:
        print(cmd)
    if set_home:
        if not env:
            env = {}
        env["HOME"] = pwd.getpwnam(user).pw_dir
    _mp_process_cmd(sudo_cmd, cwd=cwd, timeout=timeout, env=env, show_cmd=show_cmd)


def _mp_process_cmd(
    cmd: str | list,
    cwd: str | None = None,
    timeout: int | None = None,
    env: dict | None = None,
    show_cmd: bool = True,
):
    proc = mp.Process(target=_run_shell_cmd, args=(cmd, cwd, timeout, env, show_cmd))
    proc.start()
    proc.join()


def _run_shell_cmd(
    cmd: str | list,
    cwd: str | None = None,
    timeout: int | None = None,
    env: dict | None = None,
    show_cmd: bool = True,
):
    _env = dict(os.environ)
    if env:
        _env.update(env)
    if show_cmd:
        stdout = None
    else:
        stdout = DEVNULL
    run(
        cmd,
        shell=True,  # noqa: S602
        timeout=timeout,
        cwd=cwd,
        executable="/bin/bash",
        env=_env,
        stdout=stdout,
    )


def _run_cmd(
    cmd: str | list,
    cwd: str | None = None,
    timeout: int | None = None,
    env: dict | None = None,
    show_cmd: bool = True,
):
    if isinstance(cmd, str):
        cmd = shlex.split(cmd)
    if env is None:
        _env = dict(os.environ)
    else:
        _env = env
    if show_cmd:
        stdout = None
    else:
        stdout = DEVNULL
    run(
        cmd,
        shell=False,  # noqa S603
        timeout=timeout,
        cwd=cwd,
        env=_env,
        stdout=stdout,
    )


def is_current_user(user: str) -> bool:
    return pwd.getpwuid(os.getuid()).pw_name == user


def _exec_as_user(
    cmd: str | list,
    user: str,
    cwd: str | None = None,
    timeout: int | None = None,
    env: dict | None = None,
    show_cmd: bool = True,
):
    if isinstance(cmd, str):
        cmd = [cmd]
    if is_current_user(user):
        for command in cmd:
            _run_cmd(command, timeout=timeout, env=env, show_cmd=show_cmd)
    elif os.getuid() == 0 or is_current_user(NUA):
        for command in cmd:
            sudo_cmd_as_user(
                command, user, cwd=cwd, timeout=timeout, env=env, show_cmd=show_cmd
            )
    else:
        raise ValueError(f"Not allowed : _exec_as_user {user} as uid {os.getuid()}")


def mp_exec_as_user(
    cmd: str | list,
    user: str,
    *,
    cwd: str | None = None,
    env: dict | None = None,
    timeout: int | None = None,
    show_cmd: bool = True,
):
    if is_current_user(user):
        _mp_process_cmd(cmd, timeout=timeout, env=env, show_cmd=show_cmd)
    elif os.getuid() == 0 or is_current_user(NUA):
        sudo_cmd_as_user(
            cmd, user, cwd=cwd, timeout=timeout, env=env, show_cmd=show_cmd
        )
    else:
        raise ValueError(f"Not allowed : mp_exec_as_user '{user}' as uid {os.getuid()}")


def exec_as_nua(
    cmd: str | list,
    cwd: str | None = None,
    env: dict | None = None,
    timeout: int | None = None,
):
    """Exec command or list of commands as user nua."""
    _exec_as_user(cmd, NUA, cwd=cwd, timeout=timeout, env=env)


def mp_exec_as_nua(
    cmd: str | list,
    cwd: str | None = None,
    env: dict | None = None,
    timeout: int | None = None,
):
    """Exec exec_as_nua() as a python external process to allow several use of
    function without mangling ENV and UID."""
    mp_exec_as_user(cmd, NUA, cwd=cwd, timeout=timeout, env=env)


def exec_as_root_daemon(
    cmd: str,
    env: dict | None = None,
) -> mp.Process:
    """Exec a subprocess daemon as root()."""
    if os.getuid() != 0:
        if is_current_user(NUA):
            cmd = "sudo " + cmd
        else:
            raise ValueError(f"Not allowed : exec_as_root_daemon as uid {os.getuid()}")
    args = tuple(shlex.split(cmd))
    if not env:
        _env = dict(os.environ)
    else:
        _env = env
    proc = mp.Process(target=_run_cmd, args=args, kwargs={"env": _env}, daemon=True)
    proc.start()
    return proc


def mp_exec_as_postgres(
    cmd: str | list,
    cwd: str | None = None,
    env: dict | None = None,
    timeout: int | None = None,
    show_cmd: bool = True,
):
    """Exec as user 'postgres' in a python external process to allow several
    use of function without mangling ENV and UID."""
    mp_exec_as_user(
        cmd, "postgres", cwd=cwd, timeout=timeout, env=env, show_cmd=show_cmd
    )


def exec_as_root(
    cmd: str | list,
    cwd: str | None = None,
    env: dict | None = None,
    timeout: int | None = None,
    show_cmd: bool = True,
):
    """Exec command or list of commands as user root."""
    _exec_as_user(cmd, "root", cwd=cwd, timeout=timeout, env=env, show_cmd=show_cmd)


def set_nua_user() -> None:
    """Ensure user is Nua and related environment.

    Will raise OSError if fails.
    """
    nua_record = pwd.getpwnam(NUA)
    if os.getuid() != nua_record.pw_uid:
        os.setgid(nua_record.pw_gid)
        os.setuid(nua_record.pw_uid)

    os.environ["USER"] = NUA
    os.environ["HOME"] = nua_record.pw_dir
    os.environ["UID"] = str(nua_record.pw_uid)
    os.environ["SHELL"] = str(nua_record.pw_shell)
    # os.chdir(nua_record.pw_dir)



import os
from contextlib import AbstractContextManager
from pathlib import Path


# Copy/pasted from contextlib.py in Python 3.11
class chdir(AbstractContextManager):  # noqa: N801
    """Non thread-safe context manager to change the current working
    directory."""

    def __init__(self, path: Path | str):
        self.path = path
        self._old_cwd: list[str] = []

    def __enter__(self):
        try:
            self._old_cwd.append(os.getcwd())
        except OSError as e:
            print(f"Ignoring error in chdir() enter:\n{e}")
        os.chdir(self.path)

    def __exit__(self, *excinfo):
        try:
            os.chdir(self._old_cwd.pop())
        except (OSError, IndexError) as e:
            print(f"Ignoring error in chdir() exit:\n{e}")



import io
from collections.abc import Generator
from pathlib import Path
from tarfile import TarFile, TarInfo

import tomli
import yaml

from nua.lib.constants import NUA_METADATA_PATH, nua_config_names


class ArchiveSearch:
    """Utilities to search files in a container image stored as a .tar archive.

    Current usage: retrieve the nua-config.toml config used when
    building the Nua app.
    """

    def __init__(self, archive: str | Path) -> None:
        arch_path = Path(archive)
        if not arch_path.is_file():
            raise FileNotFoundError(arch_path)
        self.tar_file = TarFile(arch_path)
        self.equal_match = False

    #
    # Public API
    #
    def get_nua_config_dict(self) -> dict:
        """Return the nua-config.toml of the archive as a dict."""
        for name in nua_config_names():
            found = self.find_one(f"{NUA_METADATA_PATH}/{name}")
            if found:
                content = found[0]["content"]
                if name.endswith("toml"):
                    return tomli.loads(content)
                return yaml.safe_load(content)
        return {}

    #
    # Internal methods
    #
    def find_one(self, path_pattern: str) -> list:
        if path_pattern.startswith("/"):
            self.equal_match = True
        pattern = path_pattern.lstrip("/")
        if not pattern:
            raise ValueError("Empty pattern")
        return self._find_one(pattern)

    def _find_one(self, pattern: str) -> list:
        for tinfo in self.tar_file.getmembers():
            if tinfo.name.endswith(".tar"):
                result = list(self._sub_search(tinfo, pattern))
                if result:
                    return [result[0]]
        return []

    def _sub_search(self, tinfo: TarInfo, pattern: str) -> Generator:
        extracted = self.tar_file.extractfile(tinfo)
        if not isinstance(extracted, io.BufferedReader):
            raise StopIteration()
        with extracted as bytes_content:
            sub_tar = TarFile(fileobj=bytes_content)
            for sub_tinfo in sub_tar.getmembers():
                if (self.equal_match and sub_tinfo.name == pattern) or (
                    not self.equal_match and sub_tinfo.name.endswith(pattern)
                ):
                    result = {}
                    sub_extracted = sub_tar.extractfile(sub_tinfo)
                    if isinstance(sub_extracted, io.BufferedReader):
                        with sub_extracted as fileh:
                            result["content"] = fileh.read().decode("utf8")
                            result["path"] = sub_tinfo.name
                            result["member"] = tinfo.name
                        yield result

    def read(self, path: str | Path) -> str:
        """Return the content of a file on the archive."""
        result = self.find_one(str(path))
        return result[0]["content"].decode("utf8")



from pathlib import Path

from jinja2 import Template

from ..panic import show
from ..tool.state import verbosity


def jinja2_render_file(template: str | Path, dest: str | Path, data: dict):
    """Render a Jinja2 template file."""
    template_path = Path(template)
    if not template_path.is_file():
        raise FileNotFoundError(template_path)

    dest_path = Path(dest)
    j2_template = Template(
        template_path.read_text(encoding="utf8"), keep_trailing_newline=True
    )
    dest_path.write_text(j2_template.render(data), encoding="utf8")
    with verbosity(3):
        show("Jinja2 render template from:", template)


def jinja2_render_from_str_template(template: str, dest: str | Path, data: dict):
    """Render a Jinja2 template from a string."""
    dest_path = Path(dest)
    j2_template = Template(template, keep_trailing_newline=True)
    dest_path.write_text(j2_template.render(data), encoding="utf8")
    with verbosity(3):
        show("Jinja2 render template from string")



from datetime import datetime, timezone


def backup_date() -> str:
    """Return a string of the current datetime for backup files.

    Format: 2023-05-30-120255
    """
    return datetime.now(timezone.utc).strftime("%Y-%m-%d-%0H%0M%0S")



"""Docker scripting utils."""
import re
import string
from datetime import datetime
from functools import wraps

import docker
from docker import DockerClient
from docker.errors import APIError, BuildError, ImageNotFound
from docker.models.images import Image
from docker.utils.json_stream import json_stream

from nua.lib.panic import Abort, debug, important, print_stream, red_line, vprint
from nua.lib.tool.state import verbosity, verbosity_level

LOCAL_CONFIG = {"size_unit_MiB": False}
RE_SUCCESS = re.compile(r"(^Successfully built |sha256:)([0-9a-f]+)$")


# def vprint_log_stream(docker_log: list):
#     for line in docker_log:
#         if "stream" in line:
#             vprint("    ", line["stream"].strip())


def docker_build_log_error(func):
    @wraps(func)
    def build_log_error_wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)

        except BuildError as e:
            with verbosity(0):
                red_line("=" * 60)
                red_line("Something went wrong with image build!")
                red_line(str(e))
                red_line("=" * 60)
            raise Abort("Exiting.")

        except APIError as e:
            with verbosity(0):
                message = ["=" * 60]
                message.append("Something went wrong with image build!")
                if e.is_client_error():
                    message.append(f"{e.response.status_code} Client Error")
                elif e.is_server_error():
                    message.append(f"{e.response.status_code} Server Error")
                message.append(f"URL: {e.response.url}")
                message.append(f"Reason: {e.response.reason}")
                message.append(f"Explanation: {e.explanation}")
                message.append("=" * 60)
                for line in message:
                    red_line(line)
            raise Abort("Exiting.")

    return build_log_error_wrapper


def image_created_as_iso(image):
    return image.attrs["Created"][:19]


def docker_image_size(image):
    return image_size_repr(round(image.attrs["Size"]))


def image_size_repr(image_bytes):
    if LOCAL_CONFIG.get("size_unit_MiB"):
        return round(image_bytes / 2**20)
    return round(image_bytes / 10**6)


def size_unit():
    return "MiB" if LOCAL_CONFIG.get("size_unit_MiB") else "MB"


def docker_sanitized_name(name: str) -> str:
    """Docker valid name.

    https://docs.docker.com/engine/reference/commandline/
    tag/#extended-description
    """
    # first replace spaces per underscores
    content = "_".join(name.split())
    # then apply docjer rules
    allowed = set(string.ascii_lowercase + string.digits + ".-_")
    content = "".join([c for c in content.strip().lower() if c in allowed])
    while ("___") in content:
        content.replace("___", "__")
    for sep in ".-_":
        while content.startswith(sep):
            content = content[1:]
        while content.endswith(sep):
            content = content[:-1]
    content = content[:128]
    return content


def image_labels(reference: str) -> dict:
    image = docker_require(reference)
    if not image:
        return {}
    return image.labels


def display_docker_img(image_name: str):
    important(f"Container image for '{image_name}':")
    client = DockerClient.from_env()
    result = client.images.list(filters={"reference": image_name})
    if not result:
        red_line("No image found")
        return
    for img in result:
        display_one_docker_img(img)  # pyright: ignore


def display_one_docker_img(image: Image):
    sid = image.id.split(":")[-1][:10]  # pyright: ignore
    tags = "|".join(image.tags)
    crea = datetime.fromisoformat(image_created_as_iso(image)).isoformat(" ")
    # Note on size of image: Docker uses 10**6 for MB, not 2**20
    size = docker_image_size(image)
    vprint(f"    tags: {tags}")
    vprint(f"    id: {sid}")
    vprint(f"    size: {size}{size_unit()}")
    vprint(f"    created: {crea}")


def docker_require(reference: str) -> Image | None:
    return docker_get_locally(reference) or docker_pull(reference)


def docker_remove_locally(reference: str):
    client = DockerClient.from_env()
    try:
        image = client.images.get(reference)
        if image:
            with verbosity(3):
                debug(f"Image '{reference}' found in local Docker instance: remove it")
            client.images.remove(image=image.id, force=True, noprune=False)
    except (APIError, ImageNotFound):
        pass


def docker_get_locally(reference: str) -> Image | None:
    client = DockerClient.from_env()
    try:
        name = reference.split("/")[-1]
        image = client.images.get(name)
        if image:
            with verbosity(3):
                vprint(f"Image '{reference}' found in local Docker instance")
        return image  # pyright: ignore
    except (APIError, ImageNotFound):
        with verbosity(4):
            vprint(f"Image '{reference}' not found in local Docker instance")
        return None


def docker_pull(reference: str) -> Image | None:
    client = DockerClient.from_env()
    try:
        image = client.images.pull(reference)
        if image:
            with verbosity(3):
                vprint(f"Image '{reference}' pulled from Docker hub")
        return image  # pyright: ignore
    except (APIError, ImageNotFound):
        with verbosity(3):
            vprint(f"Image '{reference}' not found in Docker hub")
        return None


def _print_buffer_log(messages: list[str]):
    """Print messages buffered, without checking for verbosity.

    Dump retained messages when an error occured."""
    for message in messages:
        with verbosity(0):
            print_stream(message)


def _docker_stream_chunk(
    chunk: dict,
    messages_buffer: list[str],
    result: dict,
) -> None:
    if "error" in chunk:
        _print_buffer_log(messages_buffer)
        raise BuildError(chunk["error"], "")
    result["last_event"] = chunk
    message = chunk.get("stream")
    if message:
        if match := RE_SUCCESS.search(message):
            result["image_id"] = match.group(2)
        with verbosity(2):
            print_stream(message)
        if verbosity_level() < 2:
            # store message for printing full log if later an error occurs
            messages_buffer.append(message)


def docker_stream_build(
    path: str,
    tag: str,
    buildargs: dict,
    labels: dict,
) -> str:
    messages_buffer: list[str] = []
    client = docker.from_env()
    resp = client.api.build(
        path=path,
        tag=tag,
        rm=True,
        forcerm=True,
        buildargs=buildargs,
        labels=labels,
        nocache=True,
        timeout=1800,
    )
    stream = json_stream(resp)
    result: dict[str, str] = {}
    for chunk in stream:
        _docker_stream_chunk(chunk, messages_buffer, result)
    if "image_id" not in result:
        _print_buffer_log(messages_buffer)
        raise BuildError(result.get("last_event", "Unknown"), "")
    return result["image_id"]



from .version import __version__

# NB: this assumes that all the nua packages have the same version
NUA_PYTHON_TAG = f"nua-python:{__version__}"
NUA_BUILDER_TAG = f"nua-builder:{__version__}"

NUA_CONFIG_STEM = "nua-config"
NUA_CONFIG_EXT = ("json", "toml", "yaml", "yml")

# In the container:
NUA_BUILD_PATH = "/nua/build"
NUA_APP_PATH = "/nua/app"
NUA_METADATA_PATH = "/nua/metadata"
NUA_SCRIPTS_PATH = "/nua/scripts"


def nua_config_names():
    for suffix in NUA_CONFIG_EXT:
        yield f"{NUA_CONFIG_STEM}.{suffix}"



import os
from pathlib import Path

from ..shell import sh
from .apt import (
    apt_remove_lists,
    install_build_packages,
    install_package_list,
    purge_package_list,
)
from .util import append_bashrc


def npm_install(package: str, force: bool = False) -> None:
    """Install a node package globally."""
    opt = " --force" if force else ""
    cmd = f"/usr/bin/npm install -g{opt} {package}"
    sh(cmd)


# deprecated since sept 2023
# def install_nodejs_old(version: str = "16.x", keep_lists: bool = False):
#     """Install nodejs."""
#     purge_package_list("yarn npm nodejs")
#     url = f"https://deb.nodesource.com/setup_{version}"
#     target = Path("/nua") / "install_node.sh"
#     download_url(url, target)
#     for cmd in (
#         "bash /nua/install_node.sh",
#         "apt-get install -y nodejs",
#         "/usr/bin/npm update -g npm",
#         "/usr/bin/npm install -g --force yarn",
#         "/usr/bin/npm install -g --force node-gyp",
#     ):
#         sh(cmd)
#     if not keep_lists:
#         apt_remove_lists()


def install_nodejs(version: str = "16", keep_lists: bool = False):
    """Install nodejs.

    from: https://nodesource.com/
    """
    _check_supported_nodejs_version(version)
    purge_package_list("yarn npm nodejs")
    fetch_cmd = (
        "/usr/bin/curl -fsSL "
        "https://deb.nodesource.com/gpgkey/nodesource-repo.gpg.key | "
        "/usr/bin/gpg --dearmor -o /etc/apt/keyrings/nodesource.gpg"
    )
    install_cmd = (
        'echo "deb [signed-by=/etc/apt/keyrings/nodesource.gpg] '
        f'https://deb.nodesource.com/node_{version}.x nodistro main" | '
        "/usr/bin/tee /etc/apt/sources.list.d/nodesource.list"
    )
    with install_build_packages("curl gnupg", keep_lists=True):
        for cmd in (
            "mkdir -p /etc/apt/keyrings",
            fetch_cmd,
            install_cmd,
            "apt-get update",
            "apt-get install nodejs -y",
            "/usr/bin/npm install -g --force yarn",
            "/usr/bin/npm install -g --force node-gyp",
        ):
            sh(cmd)
    if not keep_lists:
        apt_remove_lists()


def _check_supported_nodejs_version(version: str) -> None:
    SUPPORTED = {"16", "18", "20"}
    if version not in SUPPORTED:
        raise RuntimeError(f"Unsupported NodeJs version {version}")


def install_nodejs_via_nvm(home: Path | str = "/nua"):
    """Install nodejs via nvm."""
    node_version_14 = "14.19.3"
    node_version = "16.18.0"
    # nvm_version = "v0.39.0"
    nvm_dir = f"{home}/.nvm"
    install_package_list("wget ", keep_lists=True)
    bashrc_modif = (
        f'export PATH="{nvm_dir}/versions/node/v{node_version}/bin/:$PATH"\n'
        f'export NVM_DIR="{nvm_dir}"\n'
        f'[ -s "$NVM_DIR/nvm.sh" ] && source "$NVM_DIR/nvm.sh"\n'
        f'[ -s "$NVM_DIR/bash_completion" ] && source "$NVM_DIR/bash_completion"'
    )
    append_bashrc(home, bashrc_modif)
    os.environ["NVM_DIR"] = ""
    os.environ["HOME"] = str(home)
    cmd = (
        f"wget -qO- https://raw.githubusercontent.com/nvm-sh/nvm/v0.39.0/install.sh "
        f"| bash  && . {nvm_dir}/nvm.sh "
        f"&& nvm install {node_version_14} "
        f"&& nvm use {node_version_14} "
        "&& npm install -g yarn "
        f"&& nvm install {node_version} "
        f"&& nvm use v{node_version} "
        "&& npm install -g yarn "
        f"&& nvm alias default v{node_version} "
        f"&& rm -rf {nvm_dir}/.cache "
    )
    environ = os.environ.copy()
    sh(cmd, env=environ)



from .apt import (
    apt_final_clean,
    apt_remove_lists,
    apt_update,
    install_build_packages,
    install_package_list,
    install_packages,
    installed_packages,
)
from .jinja import jinja2_render_file, jinja2_render_from_str_template
from .misc import install_git_source, install_meta_packages, install_source
from .nodejs import install_nodejs
from .python import (
    build_python,
    check_python_version,
    install_pip_packages,
    pip_install,
    python_package_installed,
)
from .ruby import install_ruby
from .util import (
    camel_format,
    copy_from_package,
    download_extract,
    download_url,
    kebab_format,
    replace_in,
    snake_format,
    string_in,
    to_kebab_cases,
    to_snake_cases,
)

__all__ = [
    "apt_final_clean",
    "apt_remove_lists",
    "apt_update",
    "build_python",
    "camel_format",
    "check_python_version",
    "copy_from_package",
    "download_extract",
    "download_url",
    "install_build_packages",
    "install_git_source",
    "install_meta_packages",
    "install_nodejs",
    "install_package_list",
    "install_packages",
    "install_pip_packages",
    "install_ruby",
    "install_source",
    "installed_packages",
    "jinja2_render_file",
    "jinja2_render_from_str_template",
    "kebab_format",
    "pip_install",
    "python_package_installed",
    "replace_in",
    "snake_format",
    "string_in",
    "to_kebab_cases",
    "to_snake_cases",
]



from pathlib import Path

from ..backports import chdir
from ..panic import info
from ..shell import sh
from .apt import install_package_list, installed_packages, tmp_install_package_list
from .python import pip_install
from .util import download_extract, is_local_dir


def install_meta_packages(packages: list, keep_lists: bool = False):
    """Install meta packages."""
    if "psycopg2" in packages or "postgres-client" in packages:
        info("install meta package: psycopg2")
        install_psycopg2_python(keep_lists=keep_lists)
    if "mariadb-client" in packages:
        info("install meta package: mariadb-client")
        install_mariadb_1_1_5(keep_lists=keep_lists)


def compile_openssl_1_1() -> str:
    """Compile OpenSSL 1.1.x.

    Returns:
        str: The path to the compiled OpenSSL.
    """
    soft = "openssl-1.1.1g"
    info(f"Compiling {soft}")
    dest = f"/nua/.openssl/{soft}"
    install_package_list(
        "wget pkg-config build-essential",
        keep_lists=True,
        clean=False,
    )
    with chdir("/nua"):
        sh(f"wget https://www.openssl.org/source/{soft}.tar.gz")
        sh(f"tar zxvf {soft}.tar.gz")
        with chdir(soft):
            sh(f"./config --prefix={dest} --openssldir={dest}")
            # ../test/recipes/80-test_ssl_new.t (Wstat: 256 Tests: 29 Failed: 1)
            # Failed test:  12
            # sh("make && make test && make install")
            sh("make -j4 && make install_sw && make install_ssldirs")
            sh(f"rm -fr {dest}/certs")
            sh(f"ln -s /etc/ssl/certs {dest}/certs")
        sh(f"rm -fr {soft}")
        sh(f"rm -f {soft}.tar.gz")
    return dest


def install_mariadb_1_1_5(keep_lists=True):
    """Python connector for MariaDB, since version 1.1.5post3."""
    install_package_list("libmariadb3 mariadb-client", keep_lists=keep_lists)
    with tmp_install_package_list(
        "libmariadb-dev python3-dev build-essential", keep_lists=True
    ):
        pip_install("mariadb")


def install_psycopg2_python(keep_lists: bool = True):
    """Python connector for PostgreSQL."""
    install_package_list(["libpq-dev"], keep_lists=keep_lists)
    pip_install("psycopg2-binary")


def install_source(
    url: str,
    dest_dir: str | Path,
    name: str,
    checksum: str = "",
) -> Path:
    """Guess meaning of url string and send back local path of the dowloaded
    source.

    (WIP)
    """
    if is_local_dir(url):
        return Path(url).resolve()

    return download_extract(url, dest_dir, name, checksum)


def install_git_source(url: str, branch: str, dest_dir: str | Path) -> Path:
    """Install a project from git source."""
    if dest_dir:
        path = Path(dest_dir).resolve()
    else:
        path = Path(".").resolve()
    url = url.strip()
    if url.endswith(".git"):
        url = url[:-4]
    name = url.split("/")[-1]
    cmd = f"git clone --depth 1 --branch {branch} {url} {name}"
    if "git" in installed_packages():
        with chdir(path):
            sh(cmd)
    else:
        with tmp_install_package_list("git"), chdir(path):
            sh(cmd)
    return path / name



import secrets
import string

MIN_DIGIT = 3


def gen_password(length: int = 24) -> str:
    """Generate random password.

    With ascii and digits, lenght 24, log10(62**16) ~ 43, and basic
    constraints. Added constraint for usernames: first char is letter.
    """
    alphabet = string.ascii_letters + string.digits
    while True:
        password = "".join(secrets.choice(alphabet) for i in range(length))
        if (
            password[0] in string.ascii_letters
            and any(char.islower() for char in password)
            and any(char.isupper() for char in password)
            and tuple(char.isdigit() for char in password).count(True) >= MIN_DIGIT
        ):
            break
    return password


def gen_randint() -> int:
    """Generate random integer of 64 bits."""
    return secrets.randbits(64)
